@@ -12,7 +12,7 @@ from .util.compat import SafeConfigParser
 
 class Config(object):
 
-    """Represent an Alembic configuration.
+    r"""Represent an Alembic configuration.
 
     Within an ``env.py`` script, this is available
     via the :attr:`.EnvironmentContext.config` attribute,
@@ -58,7 +58,7 @@ class Config(object):
             alembic_cfg.attributes['connection'] = connection
             command.upgrade(alembic_cfg, "head")
 
-    :param file_: name of the .ini file to open.
+    :param file\_: name of the .ini file to open.
     :param ini_section: name of the main Alembic section within the
      .ini file
     :param output_buffer: optional file-like input buffer which
@@ -212,11 +212,14 @@ class Config(object):
         """
         return os.path.join(package_dir, "templates")
 
-    def get_section(self, name):
+    def get_section(self, name, default=None):
         """Return all the configuration options from a given .ini file section
         as a dictionary.
 
         """
+        if not self.file_config.has_section(name):
+            return default
+
         return dict(self.file_config.items(name))
 
     def set_main_option(self, name, value):
@@ -422,6 +425,22 @@ class CommandLine(object):
                         help="Indicate the current revision",
                     ),
                 ),
+                "purge": (
+                    "--purge",
+                    dict(
+                        action="store_true",
+                        help="Unconditionally erase the version table "
+                        "before stamping",
+                    ),
+                ),
+                "package": (
+                    "--package",
+                    dict(
+                        action="store_true",
+                        help="Write empty __init__.py files to the "
+                        "environment and version locations",
+                    ),
+                ),
             }
             positional_help = {
                 "directory": "location of scripts directory",
